# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import types
from dune.common.hashit import hashIt
from .generator import MySimpleGenerator

from io import StringIO
from dune.generator.algorithm import run


def __fixBoundaryDOFs(dirichletValues):
    def __fixBoundaryDOFsFunc(dirichletValues, f, *args: int):
        prefixPathTypeName = "Dune::TypeTree::HybridTreePath<"
        prefixPathTypeName += ",".join(
            "Dune::index_constant<" + str(i) + ">" for i in args
        )
        prefixPathTypeName += ">"

        runCode = """
        #include <ikarus/python/dirichletvalues/dirichletvalues.hh> 
        template <typename DV>
        void fixBoundaryDofs(DV& dirichletValues, const pybind11::function& functor)
        {{
        Ikarus::Python::forwardCorrectFunction(dirichletValues, functor, 
        [&](auto&& functor_) {{ 
        return dirichletValues.fixBoundaryDOFs(functor_, {prefixPathType}{{}});
          }});
        }}
        """.format(
            prefixPathType=prefixPathTypeName
        )
        return run("fixBoundaryDofs", StringIO(runCode), dirichletValues, f)

    return __fixBoundaryDOFsFunc


def dirichletValues(basis):
    """
    @brief Creates a Dirichlet values handler for the given basis.

    @param basis: The basis.

    @return: The created Dirichlet values handler.
    """
    generator = MySimpleGenerator("DirichletValues", "Ikarus::Python")
    element_type = f"Ikarus::DirichletValues<{basis.cppTypeName},Eigen::VectorX<bool>>"

    includes = []
    includes += basis._includes
    includes += ["ikarus/utils/dirichletvalues.hh"]
    includes += ["ikarus/python/dirichletvalues/dirichletvalues.hh"]
    moduleName = "dirichletValues_" + hashIt(element_type)
    module = generator.load(
        includes=includes,
        typeName=element_type,
        moduleName=moduleName,
        dynamicAttr=True,
    )

    dv = module.DirichletValues(basis)
    dv.fixBoundaryDOFs = types.MethodType(__fixBoundaryDOFs(dv), dv)

    return dv



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later


import debug_info

debug_info.setDebugFlags()

import ikarus as iks
from ikarus import finite_elements, assembler
import numpy as np
import scipy as sp
import os

import dune.foamgrid
import dune.grid
import dune.functions


def trussTest(worldDim):
    assert worldDim == 2 or worldDim == 3
    gridDim = 1
    filedir = os.path.dirname(__file__)
    filename = os.path.join(filedir, f"auxiliaryfiles/truss{worldDim}d.msh")
    reader = (dune.grid.reader.gmsh, filename)

    grid = dune.foamgrid.foamGrid(reader, gridDim, worldDim)

    basis = iks.basis(
        grid, dune.functions.Power(dune.functions.Lagrange(order=1), worldDim)
    )

    flatBasis = basis.flat()
    d = np.zeros(len(flatBasis))

    lambdaLoad = iks.Scalar(0.0)

    fes = []

    tol = 1e-12

    if worldDim == 2:
        E = 30000
        A = 0.1
    else:
        E = 2000
        A = 1000

    trusses = finite_elements.truss(youngs_modulus=E, cross_section=A)

    for e in grid.elements:
        fes.append(finite_elements.makeFE(basis, trusses))
        fes[-1].bind(e)

    req = fes[0].createRequirement()
    req.insertParameter(lambdaLoad)

    assert req.parameter() == lambdaLoad
    req.insertGlobalSolution(d)

    d2 = req.globalSolution()

    # check that is really the same data address
    assert ("{}".format(hex(d2.__array_interface__["data"][0]))) == (
        "{}".format(hex(d.__array_interface__["data"][0]))
    )
    assert len(d2) == len(d)
    assert (d2 == d).all()

    forces = np.zeros(worldDim * 2)
    stiffness1 = np.zeros((worldDim * 2, worldDim * 2))
    stiffness2 = np.zeros((worldDim * 2, worldDim * 2))
    fes[0].calculateVector(req, iks.VectorAffordance.forces, forces)
    fes[0].calculateMatrix(req, iks.MatrixAffordance.stiffness, stiffness1)
    fes[1].calculateMatrix(req, iks.MatrixAffordance.stiffness, stiffness2)

    dirichletValues = iks.dirichletValues(flatBasis)

    def fixAllIndices(dirichletFlags, globalIndex):
        dirichletFlags[globalIndex] = True

    dirichletValues.fixBoundaryDOFs(fixAllIndices)
    if worldDim == 3:
        dirichletValues.setSingleDOF(1, True)

    denseAssembler = assembler.denseFlatAssembler(fes, dirichletValues)
    denseAssembler.bind(req, iks.AffordanceCollection.elastoStatics, iks.DBCOption.Full)

    Kdense = denseAssembler.matrix()
    forces = denseAssembler.vector()
    forces[4] += 1.0

    x = sp.linalg.solve(Kdense, -forces)

    if worldDim == 2:
        expectedDisp = (
            -0.169172906288868320
        )  # maximum vertical displacement of the center node
        expectedAxialForce = np.array([-4.592837713552849940, -4.592837713552849940])
    if worldDim == 3:
        expectedDisp = (
            -0.0008521190304919996
        )  # maximum vertical displacement of the center node
        expectedAxialForce = np.array([13.785353834067782586, -14.910220788494811472])

    assert (
        abs(abs(expectedDisp) - max(abs(x))) < tol
    ), f"The expected displacement should be {expectedDisp} but is {max(abs(x))}"

    req.insertGlobalSolution(x)

    # Test calculateAt Function
    N0 = fes[0].calculateAt(req, np.array([0.5]), "cauchyAxialForce")[0]
    N1 = fes[1].calculateAt(req, np.array([0.5]), "cauchyAxialForce")[0]
    N = np.array([N0, N1])

    for i in range(2):
        assert (
            abs(expectedAxialForce[i] - N[i]) < tol
        ), f"The expected Cauchy axial force for element {i} is {expectedAxialForce[i]} but is {N[i]}"

    # Querying for a different ResultType should result in a runtime error
    try:
        fes[0].calculateAt(req, np.array([0.5]), "PK2Stress")
    except RuntimeError:
        assert True
    else:
        assert False


if __name__ == "__main__":
    trussTest(2)
    trussTest(3)



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import debug_info

debug_info.setDebugFlags()

import ikarus as iks
import dune.grid
import dune.functions

import math
import numpy as np


def makeGrid():
    lowerLeft = []
    upperRight = []
    elements = []
    for _ in range(2):
        lowerLeft.append(-1)
        upperRight.append(1)
        elements.append(3)

    return dune.grid.structuredGrid(lowerLeft, upperRight, elements)


def testDirichletValues():
    grid = makeGrid()
    basis = iks.basis(grid, dune.functions.Power(dune.functions.Lagrange(order=1), 2))
    basis = basis.flat()

    dirichletValues = iks.dirichletValues(basis)

    assert basis.size() == dirichletValues.size
    assert basis.size() == len(dirichletValues)

    def fixOneIndex(vec, globalIndex):
        if globalIndex == 0:
            vec[globalIndex] = True

    dirichletValues.fixBoundaryDOFs(fixOneIndex)

    # This is equivalent to values[0] == True, but this syntax is discouraged by PEP
    assert dirichletValues.container[0]

    # Note that the result of localView.index(localIndex) is a multiIndex even for a flat basis, the localIndex is an int
    def fixAnotherIndexWithLocalView(vec, localIndex, localView):
        if localView.index(localIndex) == [4]:
            vec[localView.index(localIndex)] = True

        assert isinstance(localIndex, int)
        assert isinstance(localView.index(localIndex), list)
        assert isinstance(localView.index(localIndex)[0], int)

    dirichletValues.fixBoundaryDOFs(fixAnotherIndexWithLocalView)
    container = dirichletValues.container

    assert sum(container) == 2
    assert dirichletValues.fixedDOFsize == 2

    def fixTopSide(vec, localIndex, localView, intersection):
        if intersection.geometry.center[0] == 1.0:
            vec[localView.index(localIndex)] = True

        assert isinstance(localIndex, int)
        assert isinstance(localView.index(localIndex), list)
        assert isinstance(localView.index(localIndex)[0], int)

    dirichletValues.fixBoundaryDOFs(fixTopSide)

    # This assmues a structured grid
    indicesPerDirection: int = (math.sqrt(grid.size(0)) + 1) * 2
    assert dirichletValues.fixedDOFsize == 2 + indicesPerDirection

    # This checks whether container is a reference
    assert sum(container) == dirichletValues.fixedDOFsize

    # Test Subbasis
    dirichletValues2 = iks.dirichletValues(basis)

    dirichletValues2.fixBoundaryDOFs(fixOneIndex, 0)
    dirichletValues2.fixBoundaryDOFs(fixAnotherIndexWithLocalView, 0)
    dirichletValues2.fixBoundaryDOFs(fixTopSide, 0)

    assert dirichletValues2.fixedDOFsize == int(2 + indicesPerDirection / 2)

    dirichletValues2.fixBoundaryDOFs(fixTopSide, 1)
    assert dirichletValues2.fixedDOFsize == 2 + indicesPerDirection

    dirichletValues2.setSingleDOF(1, True)
    assert dirichletValues2.fixedDOFsize == 2 + indicesPerDirection + 1
    assert dirichletValues2.container[1]
    assert dirichletValues2.isConstrained(1)

    dirichletValues2.setSingleDOF((1), False)  # via MultiIndex
    assert dirichletValues2.fixedDOFsize == 2 + indicesPerDirection
    assert not dirichletValues2.isConstrained((1))  # via MultiIndex

    dirichletValues2.reset()
    assert dirichletValues2.fixedDOFsize == 0
    assert sum(dirichletValues2.container) == 0

    def fixDOFFunction(basis, vec):
        vec[:] = np.ones(dirichletValues2.size)

    dirichletValues2.fixDOFs(fixDOFFunction)
    assert dirichletValues2.fixedDOFsize == dirichletValues2.size


if __name__ == "__main__":
    testDirichletValues()



# SPDX-FileCopyrightText: 2021-2024 The Ikarus Developers mueller@ibb.uni-stuttgart.de
# SPDX-License-Identifier: LGPL-3.0-or-later

import debug_info

debug_info.setDebugFlags()

import ikarus as iks
from ikarus import materials, utils

import numpy as np
import math


def checkSizes(A, expected_rows, expected_cols):
    shape = A.shape
    if len(shape) == 2:
        num_rows, num_cols = shape
    else:
        num_rows = shape[0]
        num_cols = 1

    assert expected_rows == num_rows
    assert expected_cols == num_cols


def checkMaterial(mat, strain, nonlinear=True, size=6):
    if nonlinear:
        mat.storedEnergy(materials.StrainTags.greenLagrangian, strain)
        mat.storedEnergy(materials.StrainTags.rightCauchyGreenTensor, strain)
        S = mat.stresses(materials.StrainTags.greenLagrangian, strain)
        checkSizes(S, size, 1)
        S = mat.stresses(materials.StrainTags.rightCauchyGreenTensor, strain)
        checkSizes(S, size, 1)
        C = mat.tangentModuli(materials.StrainTags.greenLagrangian, strain)
        checkSizes(C, size, size)
        C = mat.tangentModuli(materials.StrainTags.rightCauchyGreenTensor, strain)
        checkSizes(C, size, size)
        C = mat.tangentModuli(materials.StrainTags.greenLagrangian, strain)
        checkSizes(C, size, size)
    else:
        mat.storedEnergy(materials.StrainTags.linear, strain)
        S = mat.stresses(materials.StrainTags.linear, strain)
        checkSizes(S, size, 1)
        C = mat.tangentModuli(materials.StrainTags.linear, strain)
        checkSizes(C, size, size)

    try:
        mat.storedEnergy(strain)
        assert False
    except TypeError:
        pass

    try:
        mat.stresses(materials.StrainTags.displacementGradient, strain)
        assert False
    except RuntimeError:
        pass

    try:
        mat.stresses(materials.StrainTags.deformationGradient, strain)
        assert False
    except RuntimeError:
        pass


# This checks plane stress and plane strain
def check2DReducedFullEquality(material, strainName, strain):
    strainFull = np.array(
        [
            strain[0],
            strain[1],
            (
                0
                if strainName == materials.StrainTags.linear
                or strainName == materials.StrainTags.greenLagrangian
                else 1
            ),
            0,
            0,
            strain[2],
        ]
    )

    e1 = material.storedEnergy(strainName, strainFull)
    e2 = material.storedEnergy(strainName, strain)

    assert math.isclose(e1, e2)

    s1 = material.stresses(strainName, strainFull)
    s2 = material.stresses(strainName, strain)
    assert np.allclose(s1, s2)

    C1 = material.tangentModuli(strainName, strainFull)
    C2 = material.tangentModuli(strainName, strain)
    assert np.allclose(C1, C2)


def checkShellStressReducedFullEquality(material, strainName, strain):
    strainFull = np.array(
        [
            strain[0],
            strain[1],
            (
                0
                if strainName == materials.StrainTags.linear
                or strainName == materials.StrainTags.greenLagrangian
                else 1
            ),
            strain[2],
            strain[3],
            strain[4],
        ]
    )

    e1 = material.storedEnergy(strainName, strainFull)
    e2 = material.storedEnergy(strainName, strain)

    assert math.isclose(e1, e2)

    s1 = material.stresses(strainName, strainFull)
    s2 = material.stresses(strainName, strain)
    assert np.allclose(s1, s2)

    C1 = material.tangentModuli(strainName, strainFull)
    C2 = material.tangentModuli(strainName, strain)
    assert np.allclose(C1, C2)


def checkBeamStressReducedFullEquality(material, strainName, strain):
    strainFull = np.array(
        [
            strain[0],
            (
                0
                if strainName == materials.StrainTags.linear
                or strainName == materials.StrainTags.greenLagrangian
                else 1
            ),
            (
                0
                if strainName == materials.StrainTags.linear
                or strainName == materials.StrainTags.greenLagrangian
                else 1
            ),
            strain[1],
            strain[2],
            strain[3],
        ]
    )

    e1 = material.storedEnergy(strainName, strainFull)
    e2 = material.storedEnergy(strainName, strain)

    assert math.isclose(e1, e2)

    s1 = material.stresses(strainName, strainFull)
    s2 = material.stresses(strainName, strain)
    assert np.allclose(s1, s2)

    C1 = material.tangentModuli(strainName, strainFull)
    C2 = material.tangentModuli(strainName, strain)
    assert np.allclose(C1, C2)


def checkWithStrain(strain):
    svk = iks.materials.StVenantKirchhoff(E=1000, nu=0.3)
    nh = iks.materials.NeoHooke(E=1000, nu=0.3)
    lin = iks.materials.LinearElasticity(E=1000, nu=0.3)

    if len(strain) == 6:
        checkMaterial(svk, strain)
        checkMaterial(nh, strain)
        checkMaterial(lin, strain, False)
        checkMaterial(lin.asPlaneStress(), strain, False, 3)
        checkMaterial(nh.asPlaneStress(), strain, True, 3)
        checkMaterial(svk.asPlaneStress(), strain, True, 3)

        checkMaterial(lin.asShellMaterial(), strain, False, 5)
        checkMaterial(nh.asShellMaterial(), strain, True, 5)
        checkMaterial(svk.asShellMaterial(), strain, True, 5)

        checkMaterial(lin.asBeamMaterial(), strain, False, 4)
        checkMaterial(nh.asBeamMaterial(), strain, True, 4)
        checkMaterial(svk.asBeamMaterial(), strain, True, 4)
    elif len(strain) == 3:
        checkMaterial(lin.asPlaneStress(), strain, False, 3)
        checkMaterial(nh.asPlaneStress(), strain, True, 3)
        checkMaterial(svk.asPlaneStress(), strain, True, 3)

        check2DReducedFullEquality(
            nh.asPlaneStress(), materials.StrainTags.rightCauchyGreenTensor, strain
        )
        check2DReducedFullEquality(
            nh.asPlaneStrain(), materials.StrainTags.rightCauchyGreenTensor, strain
        )
        check2DReducedFullEquality(
            svk.asPlaneStress(), materials.StrainTags.rightCauchyGreenTensor, strain
        )
        check2DReducedFullEquality(
            svk.asPlaneStrain(), materials.StrainTags.rightCauchyGreenTensor, strain
        )

        strain[0] = strain[0] - 1
        strain[1] = strain[1] - 1
        check2DReducedFullEquality(
            lin.asPlaneStress(), materials.StrainTags.linear, strain
        )
        check2DReducedFullEquality(
            lin.asPlaneStrain(), materials.StrainTags.linear, strain
        )

        check2DReducedFullEquality(
            nh.asPlaneStress(), materials.StrainTags.greenLagrangian, strain
        )
        check2DReducedFullEquality(
            nh.asPlaneStrain(), materials.StrainTags.greenLagrangian, strain
        )
        check2DReducedFullEquality(
            svk.asPlaneStress(), materials.StrainTags.greenLagrangian, strain
        )
        check2DReducedFullEquality(
            svk.asPlaneStrain(), materials.StrainTags.greenLagrangian, strain
        )

    elif len(strain) == 5:
        checkMaterial(lin.asShellMaterial(), strain, False, 5)
        checkMaterial(nh.asShellMaterial(), strain, True, 5)
        checkMaterial(svk.asShellMaterial(), strain, True, 5)

        checkShellStressReducedFullEquality(
            nh.asShellMaterial(), materials.StrainTags.rightCauchyGreenTensor, strain
        )
        checkShellStressReducedFullEquality(
            svk.asShellMaterial(), materials.StrainTags.rightCauchyGreenTensor, strain
        )

        strain[0] = strain[0] - 1
        strain[1] = strain[1] - 1
        checkShellStressReducedFullEquality(
            lin.asShellMaterial(), materials.StrainTags.linear, strain
        )
        checkShellStressReducedFullEquality(
            nh.asShellMaterial(), materials.StrainTags.greenLagrangian, strain
        )
        checkShellStressReducedFullEquality(
            svk.asShellMaterial(), materials.StrainTags.greenLagrangian, strain
        )

    elif len(strain) == 4:
        checkMaterial(lin.asBeamMaterial(), strain, False, 4)
        checkMaterial(nh.asBeamMaterial(), strain, True, 4)
        checkMaterial(svk.asBeamMaterial(), strain, True, 4)

        checkBeamStressReducedFullEquality(
            nh.asBeamMaterial(), materials.StrainTags.rightCauchyGreenTensor, strain
        )
        checkBeamStressReducedFullEquality(
            svk.asBeamMaterial(), materials.StrainTags.rightCauchyGreenTensor, strain
        )

        strain[0] = strain[0] - 1
        checkBeamStressReducedFullEquality(
            lin.asBeamMaterial(), materials.StrainTags.linear, strain
        )
        checkBeamStressReducedFullEquality(
            nh.asBeamMaterial(), materials.StrainTags.greenLagrangian, strain
        )
        checkBeamStressReducedFullEquality(
            svk.asBeamMaterial(), materials.StrainTags.greenLagrangian, strain
        )


def checkStrainTransformation():
    glVoigt = [1.2, 1.1, 0.9, 0.1, 0.2, 0.2]
    glTensor = utils.fromVoigt(glVoigt)

    c = materials.transformStrain(
        materials.StrainTags.greenLagrangian,
        materials.StrainTags.rightCauchyGreenTensor,
        glVoigt,
    )
    c2 = materials.transformStrain(
        materials.StrainTags.greenLagrangian,
        materials.StrainTags.rightCauchyGreenTensor,
        glTensor,
    )

    assert np.allclose(c, 2 * glTensor + np.identity(3))
    assert np.allclose(c2, 2 * glTensor + np.identity(3))

    gl = materials.transformStrain(
        materials.StrainTags.rightCauchyGreenTensor,
        materials.StrainTags.greenLagrangian,
        c,
    )

    assert np.allclose(gl, glTensor)
    assert np.allclose(gl, 0.5 * (c - np.identity(3)))

    gl2 = materials.transformStrain(
        materials.StrainTags.greenLagrangian, materials.StrainTags.greenLagrangian, gl
    )

    assert np.allclose(gl, gl2)


def checkVoigtTransformations():
    # Check Voigt transformations for 2D
    glVoigt = [1.2, 1.1, 0.1]
    glTensor = utils.fromVoigt(glVoigt)

    assert np.allclose(utils.toVoigt(glTensor), glVoigt)
    assert np.allclose(utils.fromVoigt(glVoigt), glTensor)

    # Check Voigt transformations for 1D
    glVoigt = [1.2]
    glTensor = utils.fromVoigt(glVoigt)

    assert np.allclose(utils.toVoigt(glTensor), glVoigt)
    assert np.allclose(utils.fromVoigt(glVoigt), glTensor)

    # Check Voigt transformations for 3D
    glVoigt = [1.2, 1.1, 0.9, 0.1, 0.2, 0.2]
    glTensor = utils.fromVoigt(glVoigt)

    assert np.allclose(utils.toVoigt(glTensor), glVoigt)
    assert np.allclose(utils.fromVoigt(glVoigt), glTensor)


def checkMaterialConstructors():
    # Check different constructors (no physical meaning)
    iks.materials.StVenantKirchhoff(E=1000, mu=500)
    iks.materials.StVenantKirchhoff(E=1000, K=500)
    iks.materials.StVenantKirchhoff(E=1000, Lambda=500)
    iks.materials.StVenantKirchhoff(K=1000, Lambda=500)
    iks.materials.StVenantKirchhoff(Lambda=1000, mu=500)

    svk1 = iks.materials.StVenantKirchhoff(E=1000, nu=0.3)
    svk2 = iks.materials.StVenantKirchhoff(nu=0.3, E=1000)

    strain = np.array([1.2, 1.1, 0.9, 0.1, 0.2, 0.2])
    s1 = svk1.stresses(materials.StrainTags.greenLagrangian, strain)
    s2 = svk2.stresses(materials.StrainTags.greenLagrangian, strain)

    assert np.allclose(s1, s2)


if __name__ == "__main__":
    strain = np.array([1.2, 1.1, 0.9, 0.1, 0.2, 0.2])
    checkWithStrain(strain)

    # check if passing plane strains is enough
    strain = np.array([1.2, 1.1, 0.1])
    checkWithStrain(strain)

    # check if passing shell strains is enough
    strain = np.array([1.2, 1.1, 0.1, 0.05, 0.2])
    checkWithStrain(strain)

    # check if passing beam strains is enough
    strain = np.array([1.2, 0.1, 0.05, 0.2])
    checkWithStrain(strain)

    checkStrainTransformation()
    checkVoigtTransformations()
    checkMaterialConstructors()
